/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef __Client_H__
#define __Client_H__

#define MAX_CONNECTIONS 256

/*
 * Set if the NX transport notified us that a flush of
 * the proxy link is needed.
 */

extern int nxagentFlush;

/*
 * Flush statistics.
 */

extern int nxagentFlushCount;
extern int nxagentFlushPending;

/*
 * The proxy can either send a ClientMessage event to
 * the agent or use the callback we provided when it
 * is going to fill all the available bandwidth. We
 * use such information to reduce the amount of data
 * produced for the proxy link until the condition is
 * ceased.
 */

extern int nxagentCongestion;

/*
 * Set when the agent can use the available bandwidth
 * to synchronize the pixmaps and the other X objects
 * that are corrupted or incomplete.
 */

extern int nxagentSynchronization;

/*
 * Proxy informs agent when it is going through a reset.
 * This is intended to help agent to detect any X error
 * arising due to incomplete or completely lost requests.
 */

extern int nxagentReset;

enum ClientHint
{
  UNKNOWN = 0,
  NXCLIENT_WINDOW,
  NXCLIENT_DIALOG
};

typedef struct _PrivClientRec
{
  int clientState;
  long clientBytes;
  enum ClientHint clientHint;

} PrivClientRec;

extern int nxagentClientPrivateIndex;

#define nxagentClientPriv(pClient) \
  ((PrivClientRec *)((pClient)->devPrivates[nxagentClientPrivateIndex].ptr))

void nxagentInitClientPrivates(ClientPtr);

#define nxagentClientAddBytes(pClient, size)	\
  (nxagentClientPriv(pClient) -> clientBytes += (size))

#define nxagentClientBytes(pClient)	\
    (nxagentClientPriv(pClient) -> clientBytes)

#define nxagentClientHint(pClient) \
    (nxagentClientPriv(pClient) -> clientHint)

#define nxagentClientIsDialog(pClient) \
    (nxagentClientHint(pClient) == NXCLIENT_DIALOG)

/*
 * The actual reason why the client
 * is sleeping.
 */

#define SleepingBySplit  1

#define nxagentNeedWakeup(client) \
               ((nxagentClientPriv(client) -> \
                      clientState) != 0)

#define nxagentNeedWakeupBySplit(client) \
               (((nxagentClientPriv(client) -> \
                      clientState) & SleepingBySplit) != 0)

void nxagentGuessClientHint(ClientPtr, Atom, char*);

/*
 * Suspend or restart the agent's
 * client.
 */

int nxagentSuspendBySplit(ClientPtr client);
int nxagentWakeupBySplit(ClientPtr client);

/*
 * Wait until the given client is
 * restarted.
 */

void nxagentWaitWakeupBySplit(ClientPtr client);

/*
FIXME: This must be moved to Drawable.h.
*/
void nxagentWaitDrawable(DrawablePtr pDrawable);

/*
 * Wakeup all the sleeping clients.
 */

void nxagentWakeupByReconnect(void);

/*
 * Reset the client state before
 * closing it down.
 */

void nxagentWakeupByReset(ClientPtr client);

/*
 * Terminate the agent at the next
 * dispatch loop.
 */

#define nxagentSetGiveUp() \
    do \
    { \
      dispatchException |= DE_TERMINATE; \
    \
      isItTimeToYield = TRUE; \
    } \
    while (0)

#endif /* __Client_H__ */
