/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXPROXY, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include <X11/X.h>
#include <X11/Xproto.h>

#include "Misc.h"
#include "Unpack.h"

#define PANIC
#define WARNING
#undef  TEST
#undef  DEBUG

#ifndef True
#define True  (1)
#define False (0)
#endif
#define Error (-1)

typedef unsigned char uint8;
typedef unsigned short uint16;
typedef unsigned int uint32;

#define CVAL(p)   (*(p++))

#define NEED_ALIGN
#undef L_ENDIAN
#undef B_ENDIAN

#if defined(__i386__) || defined(__amd64__) || defined(__x86_64__)
    #define L_ENDIAN    /* Intel 32, Intel 64, Mac-Intel and AMD64 architectures */
#elif defined(__sparc__) || defined(__APPLE__)
    #define B_ENDIAN    /* Sun or Macintosh architectures */
#endif

#ifdef NEED_ALIGN
    #ifdef L_ENDIAN
    #define CVAL2(p, v) { v = (*(p++)); v |= (*(p++)) << 8; }
    #else
    #define CVAL2(p, v) { v = (*(p++)) << 8; v |= (*(p++)); }
    #endif /* L_ENDIAN */
#else
    #define CVAL2(p, v) { v = (*((uint16*)p)); p += 2; }
#endif /* NEED_ALIGN */

#define UNROLL8(exp) { exp exp exp exp exp exp exp exp }

#define REPEAT(statement) \
{ \
  while ((count & ~0x7) && ((x+8) < width)) \
    UNROLL8( statement; count--; x++; ); \
\
  while ((count > 0) && (x < width)) { statement; count--; x++; } \
}

#define MASK_UPDATE() \
{ \
  mixmask <<= 1; \
  if (mixmask == 0) \
  { \
    mask = fom_mask ? fom_mask : CVAL(input); \
    mixmask = 1; \
  } \
}

static uint32 cvalx(uint8 **input, int Bpp);
static void setli(uint8 *input, int offset, uint32 value, int Bpp);
static uint32 getli(uint8 *input, int offset, int Bpp);

//
// This is specific of NX. From a set of RDP
// glyphs generate a number of X requests to
// display the glyphs on the drawable.
//

int UnpackRDPText(unsigned int drawable, unsigned int gcontext, int big_endian,
                      unsigned char *src_data, unsigned int src_size,
                          unsigned char *dst_data, unsigned int dst_size)
{
  unsigned char *buffer = dst_data;

  //
  // Number of glyphs.
  //

  unsigned int elements = GetULONG(src_data + 12, big_endian);

  if (elements != (src_size - 16) / 20)
  {
    #ifdef PANIC
    *logofs << "UnpackRDPText: PANIC! Text elements mismatch. "
            << (src_size - 16) / 20 << " expected " << elements
            << " found.\n" << logofs_flush;
    #endif

    cerr << "Error" << ": RDP text elements mismatch. "
         << (src_size - 16) / 20 << " expected " << elements
         << " found.\n";

    return -1;
  }

  #ifdef TEST
  *logofs << "UnpackRDPText: Using drawable " << (void *) drawable
          << " and gcontext " << (void *) gcontext << ".\n"
          << logofs_flush;
  #endif

  //
  // First X_ChangeGC sets background
  // foreground and fill style.
  //

  *buffer = (unsigned char) X_ChangeGC;

  PutUINT(6, buffer + 2, big_endian);

  PutULONG(gcontext, buffer + 4,  big_endian);

  PutULONG(GCForeground | GCBackground | GCFillStyle, buffer + 8,  big_endian);

  PutULONG(GetULONG(src_data, big_endian), buffer + 12,  big_endian);
  PutULONG(GetULONG(src_data + 4, big_endian), buffer + 16,  big_endian);
  PutULONG(GetULONG(src_data + 8, big_endian), buffer + 20,  big_endian);

  //
  // Now src_data points to the
  // first glyph element.
  //

  src_data += 16;
  buffer   += 24;

  for (unsigned int i = 0; i < elements; i++)
  {
    //
    // X_ChangeGC sets x, y and pixmap
    // for the stipple.
    //

    unsigned int pixmap, x, y, width, height;

    pixmap = GetULONG(src_data, big_endian);
    x      = GetULONG(src_data + 4, big_endian);
    y      = GetULONG(src_data + 8, big_endian);
    width  = GetULONG(src_data + 12, big_endian);
    height = GetULONG(src_data + 16, big_endian);

    *buffer = (unsigned char) X_ChangeGC;

    PutUINT(6, buffer + 2, big_endian);

    PutULONG(gcontext, buffer + 4,  big_endian);

    PutULONG(GCStipple | GCTileStipXOrigin | GCTileStipYOrigin, buffer + 8,  big_endian);

    #ifdef TEST
    *logofs << "UnpackRDPText: Building X_ChangeGC element "
            << i << " with pixmap " << (void *) pixmap
            << ".\n" << logofs_flush;
    #endif

    //
    // Stipple, tile-stipple-x-origin
    // and tile-stipple-y-origin.
    //

    PutULONG(pixmap, buffer + 12, big_endian);
    PutULONG(x,      buffer + 16, big_endian);
    PutULONG(y,      buffer + 20, big_endian);

    buffer += 24;

    //
    // Draw a rectangle using the stipple.
    //

    *buffer = (unsigned char) X_PolyFillRectangle;

    PutUINT(5, buffer + 2, big_endian);

    PutULONG(drawable, buffer + 4, big_endian);
    PutULONG(gcontext, buffer + 8, big_endian);

    PutUINT(x,      buffer + 12, big_endian);
    PutUINT(y,      buffer + 14, big_endian);
    PutUINT(width,  buffer + 16, big_endian);
    PutUINT(height, buffer + 18, big_endian);

    //
    // Go to next glyph.
    //

    buffer   += 20;
    src_data += 20;
  }

  return 1;
}

//
// The following functions are borrowed from
// the rdesktop code. They are used to unpack
// a RDP bitmap into a X image buffer.
//

int UnpackRDP(unsigned char *output, int width, int height,
                  unsigned char *input, int size, int Bpp)
{
        uint8 *end = input + size;
        uint8 *prevline = NULL, *line = NULL;
        int opcode, count, offset, isfillormix, x = width;
        int lastopcode = -1, insertmix = False, bicolour = False;
        uint8 code;
        uint32 colour1 = 0, colour2 = 0;
        uint8 mixmask, mask = 0;
        uint32 mix = 0xffffffff;
        int fom_mask = 0;
        
        while (input < end)
        {
                fom_mask = 0;
                code = CVAL(input);
                opcode = code >> 4;

                /* Handle different opcode forms */
                switch (opcode)
                {
                        case 0xc:
                        case 0xd:
                        case 0xe:
                                opcode -= 6;
                                count = code & 0xf;
                                offset = 16;
                                break;

                        case 0xf:
                                opcode = code & 0xf;
                                if (opcode < 9)
                                {
                                        count = CVAL(input);
                                        count |= CVAL(input) << 8;
                                }
                                else
                                {
                                        count = (opcode < 0xb) ? 8 : 1;
                                }
                                offset = 0;
                                break;

                        default:
                                opcode >>= 1;
                                count = code & 0x1f;
                                offset = 32;
                                break;
                }

                /* Handle strange cases for counts */
                if (offset != 0)
                {
                        isfillormix = ((opcode == 2) || (opcode == 7));

                        if (count == 0)
                        {
                                if (isfillormix)
                                        count = CVAL(input) + 1;
                                else
                                        count = CVAL(input) + offset;
                        }
                        else if (isfillormix)
                        {
                                count <<= 3;
                        }
                }

                /* Read preliminary data */
                switch (opcode)
                {
                        case 0:        /* Fill */
                                if ((lastopcode == opcode) && !((x == width) && (prevline == NULL)))
                                        insertmix = True;
                                break;
                        case 8:        /* Bicolour */
                                colour1 = cvalx(&input, Bpp);
                        case 3:        /* Colour */
                                colour2 = cvalx(&input, Bpp);
                                break;
                        case 6:        /* SetMix/Mix */
                        case 7:        /* SetMix/FillOrMix */
                                mix = cvalx(&input, Bpp);
                                opcode -= 5;
                                break;
                        case 9:        /* FillOrMix_1 */
                                mask = 0x03;
                                opcode = 0x02;
                                fom_mask = 3;
                                break;
                        case 0x0a:        /* FillOrMix_2 */
                                mask = 0x05;
                                opcode = 0x02;
                                fom_mask = 5;
                                break;

                }

                lastopcode = opcode;
                mixmask = 0;

                /* Output body */
                while (count > 0)
                {
                        if (x >= width)
                        {
                                if (height <= 0)
                                        return Error;

                                x = 0;
                                height--;

                                prevline = line;
                                line = output + height * width * Bpp;
                        }

                        switch (opcode)
                        {
                                case 0:        /* Fill */
                                        if (insertmix)
                                        {
                                                if (prevline == NULL)
                                                        setli(line, x, mix, Bpp);
                                                else
                                                        setli(line, x,
                                                              getli(prevline, x, Bpp) ^ mix, Bpp);

                                                insertmix = False;
                                                count--;
                                                x++;
                                        }

                                        if (prevline == NULL)
                                        {
                                        REPEAT(setli(line, x, 0, Bpp))}
                                        else
                                        {
                                                REPEAT(setli
                                                       (line, x, getli(prevline, x, Bpp), Bpp));
                                        }
                                        break;

                                case 1:        /* Mix */
                                        if (prevline == NULL)
                                        {
                                                REPEAT(setli(line, x, mix, Bpp));
                                        }
                                        else
                                        {
                                                REPEAT(setli
                                                       (line, x, getli(prevline, x, Bpp) ^ mix,
                                                        Bpp));
                                        }
                                        break;

                                case 2:        /* Fill or Mix */
                                        if (prevline == NULL)
                                        {
                                                REPEAT(MASK_UPDATE();
                                                       if (mask & mixmask) setli(line, x, mix, Bpp);
                                                       else
                                                       setli(line, x, 0, Bpp););
                                        }
                                        else
                                        {
                                                REPEAT(MASK_UPDATE();
                                                       if (mask & mixmask)
                                                       setli(line, x, getli(prevline, x, Bpp) ^ mix,
                                                             Bpp);
                                                       else
                                                       setli(line, x, getli(prevline, x, Bpp),
                                                             Bpp););
                                        }
                                        break;

                                case 3:        /* Colour */
                                        REPEAT(setli(line, x, colour2, Bpp));
                                        break;

                                case 4:        /* Copy */
                                        REPEAT(setli(line, x, cvalx(&input, Bpp), Bpp));
                                        break;

                                case 8:        /* Bicolour */
                                        REPEAT(if (bicolour)
                                               {
                                               setli(line, x, colour2, Bpp); bicolour = False;}
                                               else
                                               {
                                               setli(line, x, colour1, Bpp); bicolour = True;
                                               count++;}
                                        );
                                        break;

                                case 0xd:        /* White */
                                        REPEAT(setli(line, x, 0xffffffff, Bpp));
                                        break;

                                case 0xe:        /* Black */
                                        REPEAT(setli(line, x, 0, Bpp));
                                        break;

                                default:
                                        return Error;
                        }
                }
        }

        return True;
}


int UnpackRDPTo8(unsigned char *output, int width, int height,
                     unsigned char *input, int size)
{
    uint8 *end = input + size;
    uint8 *prevline = NULL, *line = NULL;
    int opcode, count, offset, isfillormix, x = width;
    int lastopcode = -1, insertmix = False, bicolour = False;
    uint8 code;
    uint8 colour1 = 0, colour2 = 0;
    uint8 mixmask, mask = 0;
    uint8 mix = 0xff;
    int fom_mask = 0;

    while (input < end)
    {
        fom_mask = 0;
        code = CVAL(input);
        opcode = code >> 4;
        /* Handle different opcode forms */
        switch (opcode)
        {
            case 0xc:
            case 0xd:
            case 0xe:
                opcode -= 6;
                count = code & 0xf;
                offset = 16;
                break;
            case 0xf:
                opcode = code & 0xf;
                if (opcode < 9)
                {
                    count = CVAL(input);
                    count |= CVAL(input) << 8;
                }
                else
                {
                    count = (opcode < 0xb) ? 8 : 1;
                }
                offset = 0;
                break;
            default:
                opcode >>= 1;
                count = code & 0x1f;
                offset = 32;
                break;
        }
        /* Handle strange cases for counts */
        if (offset != 0)
        {
            isfillormix = ((opcode == 2) || (opcode == 7));
            if (count == 0)
            {
                if (isfillormix)
                        count = CVAL(input) + 1;
                else
                        count = CVAL(input) + offset;
            }
            else 
            if (isfillormix)
            {
                count <<= 3;
            }
        }
        /* Read preliminary data */
        switch (opcode)
        {
            case 0:        /* Fill */
                if ((lastopcode == opcode) && !((x == width) && (prevline == NULL)))
                insertmix = True;
                break;
            
            case 8:        /* Bicolour */
                colour1 = CVAL(input);
            
            case 3:        /* Colour */
                colour2 = CVAL(input);
                break;
            
            case 6:        /* SetMix/Mix */
            
            case 7:        /* SetMix/FillOrMix */
                mix = CVAL(input);
                opcode -= 5;
                break;
            
            case 9:        /* FillOrMix_1 */
                mask = 0x03;
                opcode = 0x02;
                fom_mask = 3;
                break;
            
            case 0x0a:        /* FillOrMix_2 */
                mask = 0x05;
                opcode = 0x02;
                fom_mask = 5;
                break;
        }
        lastopcode = opcode;
        mixmask = 0;
        /* Output body */
        while (count > 0)
        {
            if (x >= width)
            {
                if (height <= 0)
                        return False;
                x = 0;
                height--;
                prevline = line;
                line = output + height * width;
            }
            switch (opcode)
            {
                
                case 0:        /* Fill */
                    if (insertmix)
                    {
                        if (prevline == NULL)
                            line[x] = mix;
                        else
                            line[x] = prevline[x] ^ mix;
                        insertmix = False;
                        count--;
                        x++;
                    }
                    if (prevline == NULL)
                    {
                        REPEAT(line[x] = 0)
                    }
                    else
                    {
                        REPEAT(line[x] = prevline[x])
                    }
                    break;
                
                case 1:        /* Mix */
                    if (prevline == NULL)
                    {
                        REPEAT(line[x] = mix)
                    }
                    else
                    {
                        REPEAT(line[x] = prevline[x] ^ mix)
                    }
                    break;
                
                case 2:        /* Fill or Mix */
                    if (prevline == NULL)
                    {
                        REPEAT(
                                MASK_UPDATE();
                                if (mask & mixmask)
                                    line[x] = mix;
                                else
                                    line[x] = 0;
                              )
                    }
                    else
                    {
                        REPEAT(
                                MASK_UPDATE();
                                if (mask & mixmask)
                                    line[x] = prevline[x] ^ mix;
                                else
                                    line[x] = prevline[x];
                               )
                    }
                    break;
                
                case 3:        /* Colour */
                    REPEAT(line[x] = colour2)
                break;
                
                case 4:        /* Copy */
                    REPEAT(line[x] = CVAL(input))
                break;
                
                case 8:        /* Bicolour */
                    REPEAT(
                            if (bicolour)
                            {
                                line[x] = colour2;
                                bicolour = False;
                            }
                            else
                            {
                                line[x] = colour1;
                                bicolour = True; count++;
                            }
                          )
                break;
                
                case 0xd:        /* White */
                    REPEAT(line[x] = 0xff)
                break;
                
                case 0xe:        /* Black */
                    REPEAT(line[x] = 0)
                break;
                
                default:
                    #ifdef PANIC
                    *logofs << "UnpackRDPTo8: PANIC! Unimplemented bitmap opcode "
                                << (void *) opcode << ".\n" << logofs_flush;
                    #endif
                    cerr << "Error" << ": Unimplemented bitmap opcode "
                             << (void *) opcode << " in RDP image.\n";
                    return False;
                }
                
        }
    }

    return True;
}                     

int UnpackRDPTo16(unsigned char *output, int width, int height,
                      unsigned char *input, int size)
{
    uint8 *end = input + size;
    uint16 *prevline = NULL, *line = NULL;
    int opcode = -1;
    int count, offset, isfillormix, x = width;
    int lastopcode = -1, insertmix = False, bicolour = False;
    uint8 code;
    uint16 colour1 = 0, colour2 = 0;
    uint8 mixmask, mask = 0;
    uint16 mix = 0xffff;
    int fom_mask = 0;
        
    while (input < end)
    {
        fom_mask = 0;
        code = CVAL(input);
        opcode = code >> 4;
        /* Handle different opcode forms */
        switch (opcode)
        {
            case 0xc:
        
            case 0xd:
            
            case 0xe:
                opcode -= 6;
                count = code & 0xf;
                offset = 16;
            break;
        
            case 0xf:
                opcode = code & 0xf;
                if (opcode < 9)
                {
                    count = CVAL(input);
                    count |= CVAL(input) << 8;
                }
                else
                {
                    count = (opcode < 0xb) ? 8 : 1;
                }
                offset = 0;
            break;
            
            default:
                opcode >>= 1;
                count = code & 0x1f;
                offset = 32;
            break;
        }
        /* Handle strange cases for counts */
        if (offset != 0)
        {
            isfillormix = ((opcode == 2) || (opcode == 7));
            if (count == 0)
            {
                if (isfillormix)
                    count = CVAL(input) + 1;
                else
                    count = CVAL(input) + offset;
            }
            else 
            if (isfillormix)
            {
                count <<= 3;
            }
        }
        /* Read preliminary data */
        switch (opcode)
        {
            case 0:        /* Fill */
                if ((lastopcode == opcode) && !((x == width) && (prevline == NULL)))
                insertmix = True;
            break;
        
            case 8:        /* Bicolour */
                CVAL2(input, colour1);
        
            case 3:        /* Colour */
                CVAL2(input, colour2);
            break;
        
            case 6:        /* SetMix/Mix */
            
            case 7:        /* SetMix/FillOrMix */
                CVAL2(input, mix);
                opcode -= 5;
            break;
        
            case 9:        /* FillOrMix_1 */
                mask = 0x03;
                opcode = 0x02;
                fom_mask = 3;
            break;
            
            case 0x0a:        /* FillOrMix_2 */
                mask = 0x05;
                opcode = 0x02;
                fom_mask = 5;
            break;
        }
        
        lastopcode = opcode;
        mixmask = 0;
        /* Output body */
        while (count > 0)
        {
            if (x >= width)
            {
                if (height <= 0)
                {
                    #ifdef PANIC
                      *logofs << "UnpackRDPTo16: PANIC! Bad height value " << height
                            << " in RDP image.\n" << logofs_flush;
                    #endif
                    cerr << "Error" << ": Bad height value " << height
                                        << " in RDP image.\n";
                    return Error;
                }
                x = 0;
                height--;
                prevline = line;
                line = ((uint16 *) output) + height * width;
            }
            switch (opcode)
            {
                case 0:        /* Fill */
                    if (insertmix)
                    {
                        if (prevline == NULL)
                            line[x] = mix;
                        else
                            line[x] = prevline[x] ^ mix;
                        insertmix = False;
                        count--;
                        x++;
                    }
                    if (prevline == NULL)
                    {
                        REPEAT(line[x] = 0)
                    }
                    else
                    {
                        REPEAT(line[x] = prevline[x])
                    }
                break;
                
                case 1:        /* Mix */
                    if (prevline == NULL)
                    {
                        REPEAT(line[x] = mix)
                    }
                        else
                    {
                        REPEAT(line[x] = prevline[x] ^ mix)
                    }
                break;
                
                case 2:        /* Fill or Mix */
                    if (prevline == NULL)
                    {
                        REPEAT
                        (
                            MASK_UPDATE();
                            if (mask & mixmask)
                                line[x] = mix;
                            else
                                line[x] = 0;
                        )
                    }
                    else
                    {
                        REPEAT
                        (
                            MASK_UPDATE();
                            if (mask & mixmask)
                                line[x] = prevline[x] ^ mix;
                            else
                                line[x] = prevline[x];
                        )
                    }
                break;
                
                case 3:        /* Colour */
                    REPEAT(line[x] = colour2)
                break;
                
                case 4:        /* Copy */
                    REPEAT(CVAL2(input, line[x]))
                break;
                
                case 8:        /* Bicolour */
                    REPEAT
                    (
                        if (bicolour)
                        {
                            line[x] = colour2;
                            bicolour = False;
                        }
                        else
                        {
                            line[x] = colour1;
                            bicolour = True;
                            count++;
                        }
                    )
                break;
                
                case 0xd:        /* White */
                    REPEAT(line[x] = 0xffff)
                break;
                
                case 0xe:        /* Black */
                    REPEAT(line[x] = 0)
                break;
                
                default:
                    #ifdef PANIC
                    *logofs << "UnpackRDPTo16: PANIC! Unimplemented bitmap opcode "
                                << (void *) opcode << ".\n" << logofs_flush;
                    #endif
                    cerr << "Error" << ": Unimplemented bitmap opcode "
                            << (void *) opcode << " in RDP image.\n";
                return Error;
            }
        }
    }

    return True;
}

int UnpackRDPTo24(unsigned char *output, int width, int height,
                      unsigned char *input, int size)
{
    uint8 *end = input + size;
    uint8 *prevline = NULL, *line = NULL;
    int opcode, count, offset, isfillormix, x = width;
    int lastopcode = -1, insertmix = False, bicolour = False;
    uint8 code;
    uint8 colour1[3] = {0, 0, 0}, colour2[3] = {0, 0, 0};
    uint8 mixmask, mask = 0;
    uint8 mix[3] = {0xff, 0xff, 0xff};
    int fom_mask = 0;

    while (input < end)
    {
        fom_mask = 0;
        code = CVAL(input);
        opcode = code >> 4;
        /* Handle different opcode forms */
        switch (opcode)
        {
            case 0xc:
            
            case 0xd:
            
            case 0xe:
                opcode -= 6;
                count = code & 0xf;
                offset = 16;
            break;
            
            case 0xf:
                opcode = code & 0xf;
                if (opcode < 9)
                {
                    count = CVAL(input);
                    count |= CVAL(input) << 8;
                }
                else
                {
                    count = (opcode < 0xb) ? 8 : 1;
                }
                offset = 0;
            break;
            
            default:
                opcode >>= 1;
                count = code & 0x1f;
                offset = 32;
            break;
        }
        /* Handle strange cases for counts */
        if (offset != 0)
        {
            isfillormix = ((opcode == 2) || (opcode == 7));
            if (count == 0)
            {
                if (isfillormix)
                    count = CVAL(input) + 1;
                else
                    count = CVAL(input) + offset;
            }
            else 
            if (isfillormix)
            {
                count <<= 3;
            }
        }
        /* Read preliminary data */
        switch (opcode)
        {
            case 0:        /* Fill */
                if ((lastopcode == opcode) && !((x == width) && (prevline == NULL)))
                insertmix = True;
            break;
        
            case 8:        /* Bicolour */
                colour1[0] = CVAL(input);
                colour1[1] = CVAL(input);
                colour1[2] = CVAL(input);
            
            case 3:        /* Colour */
                colour2[0] = CVAL(input);
                colour2[1] = CVAL(input);
                colour2[2] = CVAL(input);
            break;
            
            case 6:        /* SetMix/Mix */
            
            case 7:        /* SetMix/FillOrMix */
                mix[0] = CVAL(input);
                mix[1] = CVAL(input);
                mix[2] = CVAL(input);
                opcode -= 5;
            break;
        
            case 9:        /* FillOrMix_1 */
                mask = 0x03;
                opcode = 0x02;
                fom_mask = 3;
            break;
        
            case 0x0a:        /* FillOrMix_2 */
                mask = 0x05;
                opcode = 0x02;
                fom_mask = 5;
            break;
        }
        lastopcode = opcode;
        mixmask = 0;
        /* Output body */
        while (count > 0)
        {
            if (x >= width)
            {
                if (height <= 0)
                {
                    #ifdef PANIC
                      *logofs << "UnpackRDPTo24: PANIC! Bad height value " << height
                            << " in RDP image.\n" << logofs_flush;
                    #endif
                    cerr << "Error" << ": Bad height value " << height
                                        << " in RDP image.\n";

                    return Error;
                }
                x = 0;
                height--;
                prevline = line;
                line = output + height * (width * 3);
            }
            switch (opcode)
            {
                case 0:        /* Fill */
                    if (insertmix)
                    {
                        if (prevline == NULL)
                        {
                            line[x * 3] = mix[0];
                            line[x * 3 + 1] = mix[1];
                            line[x * 3 + 2] = mix[2];
                        }
                        else
                        {
                            line[x * 3] =
                            prevline[x * 3] ^ mix[0];
                            line[x * 3 + 1] =
                            prevline[x * 3 + 1] ^ mix[1];
                            line[x * 3 + 2] =
                            prevline[x * 3 + 2] ^ mix[2];
                        }
                        insertmix = False;
                        count--;
                        x++;
                    }
                    if (prevline == NULL)
                    {
                        REPEAT
                        (
                            line[x * 3] = 0;
                            line[x * 3 + 1] = 0;
                            line[x * 3 + 2] = 0;
                        )
                    }
                    else
                    {
                        REPEAT
                        (
                            line[x * 3] = prevline[x * 3];
                            line[x * 3 + 1] = prevline[x * 3 + 1];
                            line[x * 3 + 2] = prevline[x * 3 + 2];
                        )
                    }
                break;
                
                case 1:        /* Mix */
                    if (prevline == NULL)
                    {
                        REPEAT
                        (
                            line[x * 3] = mix[0];
                            line[x * 3 + 1] = mix[1];
                            line[x * 3 + 2] = mix[2];
                        )
                    }
                    else
                    {
                        REPEAT
                        (
                            line[x * 3] =
                            prevline[x * 3] ^ mix[0];
                            line[x * 3 + 1] =
                            prevline[x * 3 + 1] ^ mix[1];
                            line[x * 3 + 2] =
                            prevline[x * 3 + 2] ^ mix[2];
                        )
                    }
                break;
                
                case 2:        /* Fill or Mix */
                    if (prevline == NULL)
                    {
                        REPEAT
                        (
                            MASK_UPDATE();
                            if (mask & mixmask)
                            {
                                line[x * 3] = mix[0];
                                line[x * 3 + 1] = mix[1];
                                line[x * 3 + 2] = mix[2];
                            }
                            else
                            {
                                line[x * 3] = 0;
                                line[x * 3 + 1] = 0;
                                line[x * 3 + 2] = 0;
                            }
                        )
                    }
                    else
                    {
                        REPEAT
                        (
                            MASK_UPDATE();
                            if (mask & mixmask)
                            {
                                line[x * 3] =  prevline[x * 3] ^ mix [0];
                                line[x * 3 + 1] = prevline[x * 3 + 1] ^ mix [1];
                                line[x * 3 + 2] = prevline[x * 3 + 2] ^ mix [2];
                            }
                            else
                            {
                                line[x * 3] = prevline[x * 3];
                                line[x * 3 + 1] = prevline[x * 3 + 1];
                                line[x * 3 + 2] = prevline[x * 3 + 2];
                            }
                        )
                    }
                break;
                
                case 3:        /* Colour */
                    REPEAT
                    (
                        line[x * 3] = colour2 [0];
                        line[x * 3 + 1] = colour2 [1];
                        line[x * 3 + 2] = colour2 [2];
                    )
                break;
                
                case 4:        /* Copy */
                    REPEAT
                    (
                        line[x * 3] = CVAL(input);
                        line[x * 3 + 1] = CVAL(input);
                        line[x * 3 + 2] = CVAL(input);
                    )
                break;
                
                case 8:        /* Bicolour */
                    REPEAT
                    (
                        if (bicolour)
                        {
                            line[x * 3] = colour2[0];
                            line[x * 3 + 1] = colour2[1];
                            line[x * 3 + 2] = colour2[2];
                            bicolour = False;
                        }
                        else
                        {
                            line[x * 3] = colour1[0];
                            line[x * 3 + 1] = colour1[1];
                            line[x * 3 + 2] = colour1[2];
                            bicolour = True;
                            count++;
                        }
                    )
                break;
                
                case 0xd:        /* White */
                    REPEAT
                    (
                        line[x * 3] = 0xff;
                        line[x * 3 + 1] = 0xff;
                        line[x * 3 + 2] = 0xff;
                    )
                break;
                
                case 0xe:        /* Black */
                    REPEAT
                    (
                        line[x * 3] = 0;
                        line[x * 3 + 1] = 0;
                        line[x * 3 + 2] = 0;
                    )
                break;
                
                default:
                    #ifdef PANIC
                    *logofs << "UnpackRDPTo24: PANIC! Unimplemented bitmap opcode "
                                << (void *) opcode << ".\n" << logofs_flush;
                    #endif
                    cerr << "Error" << ": Unimplemented bitmap opcode "
                        << (void *) opcode << " in RDP image.\n";

                    return Error;
            }
        }
    }

    return True;
}

uint32 cvalx(uint8 **input, int Bpp)
{
        uint32 rv = 0;
        memcpy(&rv, *input, Bpp);
        *input += Bpp;
        return rv;
}

void setli(uint8 *input, int offset, uint32 value, int Bpp)
{
        input += offset * Bpp;
        memcpy(input, &value, Bpp);
}

uint32 getli(uint8 *input, int offset, int Bpp)
{
        uint32 rv = 0;
        input += offset * Bpp;
        memcpy(&rv, input, Bpp);
        return rv;
}
